@@ -1,5 +1,5 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997, 2005-2007, 2009-2011, 2014-2016 Sergey Poznyakoff
+   Copyright (C) 1997-2021 Sergey Poznyakoff
  
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -47,11 +47,11 @@ void maybe_parm_list(int *parm_cnt_return);
 void call(char*, int);
 void reference(char*, int);
 
-int level;                  /* Current nesting level */
-Symbol *caller;             /* Current caller */
-struct obstack text_stk;    /* Obstack for composing declaration line */
+static int level;                  /* Current nesting level */
+static Symbol *caller;             /* Current caller */
+static struct obstack text_stk;    /* Obstack for composing declaration line */
 
-int parm_level;             /* Parameter declaration nesting level */
+static int parm_level;             /* Parameter declaration nesting level */
 
 typedef struct {
      int type;
@@ -61,10 +61,10 @@ typedef struct {
 
 typedef int Stackpos[1];
 
-TOKSTK tok;
-TOKSTK *token_stack;
-int tos;
-int curs;
+static TOKSTK tok;
+static TOKSTK *token_stack;
+static int tos;
+static int curs;
 int token_stack_length = 64;
 int token_stack_increase = 32;
 static int need_space;
@@ -236,7 +236,7 @@ tokdel(int beg, int end)
      if (end >= beg) {
 	  if (end < tos)
 	       memmove(token_stack + beg, token_stack + end + 1,
-		       (end - beg + 1) * sizeof(token_stack[0]));
+		       (tos - end - 1) * sizeof(token_stack[0]));
 	  tos -= (end - beg + 1);
      }
 }
@@ -379,6 +379,20 @@ save_token(TOKSTK *tokptr)
 	  obstack_1grow(&text_stk, tokptr->type);
 	  need_space = 0;
 	  break;
+     case LBRACE:
+     case LBRACE0:
+	  if (need_space) 
+	       obstack_1grow(&text_stk, ' ');
+	  obstack_1grow(&text_stk, '{');
+	  need_space = 1;
+	  break;
+     case RBRACE:
+     case RBRACE0:
+	  if (need_space) 
+	       obstack_1grow(&text_stk, ' ');
+	  obstack_1grow(&text_stk, '}');
+	  need_space = 1;
+	  break;
      case OP:
 	  obstack_1grow(&text_stk, ' ');
 	  obstack_grow(&text_stk, tokptr->token, strlen(tokptr->token));
@@ -406,6 +420,12 @@ undo_save_stack()
      save_end = -1;
 }
 
+int
+save_stack_is_empty()
+{
+     return save_end <= 0;
+}
+	  
 char *
 finish_save_stack(char *name)
 {
@@ -574,6 +594,14 @@ expression()
 	  switch (tok.type) {
 	  case ';':
 	       return;
+	  case PARM_WRAPPER:
+	       if (skip_balanced('(', ')', 0) == -1) {
+		    file_error(_("unexpected end of file in expression"),
+			       NULL);
+		    return;
+	       }
+	       putback();
+	       break;
 	  case LBRACE:
 	  case LBRACE0:
 	  case RBRACE:
@@ -599,9 +627,17 @@ expression()
 		    parens_lev++;
 	       } else {
 		    reference(name, line);
+
+		    /* MEMBER_OF can be preceded by a closing paren, as in
+		       (*a)->b
+		    */
+		    while (parens_lev > 0 && tok.type == ')') {
+			 parens_lev--;
+			 nexttoken();
+		    }
+
 		    if (tok.type == MEMBER_OF) {
-			 while (tok.type == MEMBER_OF)
-			      nexttoken();
+			 nexttoken();
 		    } else {
 			 putback();
 		    }
@@ -609,16 +645,30 @@ expression()
 	       break;
 	  case '(':
 	       /* maybe typecast */
-	       if (nexttoken() == TYPE)
-		    skip_to(')');
-	       else {
-		    putback();
+	       if (nexttoken() == TYPE || tok.type == STRUCT) {
+		    if (skip_balanced('(', ')', 1) == -1) {
+			 file_error(_("unexpected end of file in expression"),
+				    NULL);
+			 return;
+		    }
+		    if (tok.type == LBRACE || tok.type == LBRACE0) {
+			 if (skip_balanced('{', '}', 1) == -1) {
+			      file_error(_("unexpected end of file in expression"),
+					 NULL);
+			      return;
+			 }
+		    }
+	       } else {
 		    parens_lev++;
 	       }
+	       putback();
 	       break;
 	  case ')':
 	       parens_lev--;
 	       break;
+	  case MEMBER_OF:
+	       nexttoken();
+	       break;
 	  }
 	  nexttoken();
      }
@@ -654,7 +704,8 @@ parse_function_declaration(Ident *ident, int parm)
      case LBRACE:
 	  if (ident->name) {
 	       caller = lookup(ident->name);
-	       if (caller && caller->storage == AutoStorage)
+	       if (caller &&
+		   (caller->storage == AutoStorage || caller->flag == symbol_target))
 		    caller = NULL;
 	       func_body();
 	  }
@@ -668,9 +719,6 @@ parse_function_declaration(Ident *ident, int parm)
 int
 fake_struct(Ident *ident)
 {
-     Stackpos sp;
-     
-     mark(sp);
      ident->type_end = -1;
      if (tok.type == STRUCT) {
 	  if (nexttoken() == IDENTIFIER) {
@@ -678,15 +726,8 @@ fake_struct(Ident *ident)
 	  }
 	  putback();
 	  skip_struct();
-	  if (tok.type == IDENTIFIER || tok.type == MODIFIER) {
-	       int pos = curs-1;
-	       restore(sp);
-	       if (ident->type_end == -1) {
-		    /* there was no tag. Insert { ... } */
-		    tokdel(curs, pos - 1);
-		    tokins(curs, IDENTIFIER, tok.line, "{ ... }");
-		    debugtoken(&tok, "modified stack");
-	       }
+	  if (tok.type == IDENTIFIER || tok.type == MODIFIER || tok.type == QUALIFIER) {
+	       putback();
 	  } else if (tok.type == '(')
 	       return 0;
 	  else if (tok.type != ';')
@@ -712,9 +753,9 @@ parse_variable_declaration(Ident *ident, int parm)
 	  while (tok.type == MODIFIER || tok.type == QUALIFIER)
 	       nexttoken();
 	  if (tok.type == IDENTIFIER) {
-	       int pos = curs-1;
-	       restore(sp);
 	       if (ident->type_end == -1) {
+		    int pos = curs-1;
+		    restore(sp);
 		    /* there was no tag. Insert { ... } */
 		    tokdel(curs, pos - 1);
 		    tokins(curs, IDENTIFIER, tok.line, "{ ... }");
@@ -755,7 +796,6 @@ parse_variable_declaration(Ident *ident, int parm)
 	  else
 	       expression();
 	  goto select;
-	  break;
      case LBRACE0:
      case LBRACE:
 	  func_body();
@@ -820,8 +860,10 @@ skip_struct()
      }
 
      while (tok.type == PARM_WRAPPER) {
-	  if (skip_balanced('(', ')', 0) == -1)
+	  if (skip_balanced('(', ')', 0) == -1) {
 	      file_error(_("unexpected end of file in struct"), NULL);
+	      return;
+	  }
      }
 }
 
@@ -872,6 +914,7 @@ dcl(Ident *idptr)
 			       NULL);
 		    return 1;
 	       }
+	       putback();	       
 	  } else if (tok.type == IDENTIFIER) {
 	       int type;
 	       
@@ -1043,7 +1086,7 @@ func_body()
 	       parse_declaration(&ident, 0);
 	       break;
 	  case LBRACE0:
-	  case '{':
+	  case LBRACE:
 	       level++;
 	       break;
 	  case RBRACE0:
@@ -1057,7 +1100,7 @@ func_body()
 	       }
 	       /* else: */
 	       /* FALLTHRU */
-	  case '}':
+	  case RBRACE:
 	       delete_autos(level);
 	       level--;
 	       break;
@@ -1065,10 +1108,10 @@ func_body()
 	       if (verbose)
 		    file_error(_("unexpected end of file in function body"),
 			       NULL);
-	       caller = NULL;
-	       return;
+	       goto end;
 	  }
      }
+end:
      caller = NULL;
 }
 
@@ -1080,6 +1123,7 @@ get_knr_args(Ident *ident)
      Ident id;
 
      switch (tok.type) {
+     case QUALIFIER:
      case IDENTIFIER:
      case TYPE:
      case STRUCT:
@@ -1098,6 +1142,7 @@ get_knr_args(Ident *ident)
 		    stop = 1;
 		    break;
 	       case TYPE:
+	       case QUALIFIER:
 	       case IDENTIFIER:
 	       case STRUCT:
 		    putback();
@@ -1146,7 +1191,8 @@ declare(Ident *ident, int maybe_knr)
 	  && (!maybe_knr || get_knr_args(ident) == 0)
 	  && !(tok.type == LBRACE || tok.type == LBRACE0 || tok.type == TYPE
 	       || tok.type == PARM_WRAPPER))
-	 || (ident->parmcnt < 0 && ident->storage == ExplicitExternStorage)) {
+	 || (ident->parmcnt < 0 && ident->storage == ExplicitExternStorage)
+	 || save_stack_is_empty()) {
 	  undo_save_stack();
 	  /* add_external()?? */
 	  return;
@@ -1217,11 +1263,18 @@ get_symbol(char *name)
      
      if (sp) {
 	  for (; sp; sp = sp->next) {
-	       if (sp->type == SymIdentifier && strcmp(sp->name, name) == 0)
+	       if (sp->flag == symbol_start ||
+		   sp->flag == symbol_target ||
+		   sp->type == SymIdentifier)
 		    break;
 	  }
-	  if (sp)
+	  if (sp) {
+	       if ((sp->flag == symbol_start || sp->flag == symbol_target)
+		   && sp->type == SymUndefined) {
+		    init_ident(sp, ExternStorage);
+	       }
 	       return sp;
+	  }
      }
      return install_ident(name, ExternStorage);
 }
@@ -1277,3 +1330,14 @@ reference(char *name, int line)
      }
 }
 
+/* If the caller is a static symbol, reset it to NULL.  This function
+   is called as a part of cleanup routine at the end of compilation
+   module. */
+void
+reset_static_caller(void)
+{
+     if (caller &&
+	 (caller->storage == StaticStorage || caller->flag == symbol_local))
+	  caller = NULL;
+}
+
